"""
Sobol' total order estimators module
"""

from abc import ABC, abstractmethod

import numpy as np


class SobolEstimator(ABC):
    """
    Base class for Sobol' total order estimators.
    """

    @staticmethod
    def masks_dim(masks):
        """
        Deduce the number of dimensions using the sampling masks.

        Parameters
        ----------
        masks
          Low resolution masks (before upsampling) used, one for each output.

        Returns
        -------
        nb_dim
          The number of dimensions under study according to the masks.
        """
        nb_dim = np.prod(masks.shape[1:])
        return nb_dim

    @staticmethod
    def split_abc(outputs, nb_design, nb_dim):
        """
        Split the outputs values into the 3 sampling matrices A, B and C.

        Parameters
        ----------
        outputs
          Model outputs for each sample point of matrices A, B and C (in order).
        nb_design
          Number of points for matrices A (the same as B).
        nb_dim
          Number of dimensions to estimate.

        Returns
        -------
        a
          The results for the sample points in matrix A.
        b
          The results for the sample points in matrix A.
        c
          The results for the sample points in matrix C.
        """
        sampling_a = outputs[:nb_design]
        sampling_b = outputs[nb_design:nb_design*2]
        replication_c = np.array([outputs[nb_design*2 + nb_design*i:nb_design*2 + nb_design*(i+1)]
                      for i in range(nb_dim)])
        return sampling_a, sampling_b, replication_c

    @staticmethod
    def post_process(stis, masks):
        """
        Post processing ops on the indices before sending them back. Makes sure the data
        format and shape is correct.

        Parameters
        ----------
        stis
          Total order Sobol' indices, one for each dimensions.
        masks
            Low resolution masks (before upsampling) used, one for each output.

        Returns
        -------
        stis
          Total order Sobol' indices after post processing.
        """
        stis = np.array(stis, np.float32)
        return stis.reshape(masks.shape[1:])

    @abstractmethod
    def __call__(self, masks, outputs, nb_design):
        """
        Compute the Sobol' total order indices according to the Jansen algorithm.

        Ref. Jansen, M., Analysis of variance designs for model output (1999)
        https://www.sciencedirect.com/science/article/abs/pii/S0010465598001544

        Parameters
        ----------
        masks
          Low resolution masks (before upsampling) used, one for each output.
        outputs
          Model outputs associated to each masks. One for each sample point of
          matrices A, B and C (in order).
        nb_design
          Number of points for matrices A (the same as B).

        Returns
        -------
        sti: ndarray
          Total order Sobol' indices, one for each dimensions.
        """
        raise NotImplementedError()


class JansenEstimator(SobolEstimator):
    """
    Jansen estimator for total order Sobol' indices.

    Ref. Jansen, M., Analysis of variance designs for model output (1999)
    https://www.sciencedirect.com/science/article/abs/pii/S0010465598001544
    """

    def __call__(self, masks, outputs, nb_design):
        """
        Compute the Sobol' total order indices according to the Jansen algorithm.

        Parameters
        ----------
        masks
          Low resolution masks (before upsampling) used, one for each output.
        outputs
          Model outputs associated to each masks. One for each sample point of
          matrices A, B and C (in order).
        nb_design
          Number of points for matrices A (the same as B).

        Returns
        -------
        sti
          Total order Sobol' indices, one for each dimensions.
        """
        nb_dim = self.masks_dim(masks)
        sampling_a, _, replication_c = self.split_abc(outputs, nb_design, nb_dim)

        mu_a = np.mean(sampling_a)
        var = np.sum([(v - mu_a)**2 for v in sampling_a]) / (len(sampling_a) - 1)

        stis = [
            np.sum((sampling_a - replication_c[i])**2.0) / (2 * nb_design * var)
            for i in range(nb_dim)
        ]

        return self.post_process(stis, masks)


class HommaEstimator(SobolEstimator):
    """
    Homma estimator for total order Sobol' indices.

    Ref. Homma & al., Importance measures in global sensitivity analysis of nonlinear models (1996)
    https://www.sciencedirect.com/science/article/abs/pii/0951832096000026
    """

    def __call__(self, masks, outputs, nb_design):
        """
        Compute the Sobol' total order indices according to the Homma-Saltelli algorithm.

        Parameters
        ----------
        masks
          Low resolution masks (before upsampling) used, one for each output.
        outputs
          Model outputs associated to each masks. One for each sample point of
          matrices A, B and C (in order).
        nb_design
          Number of points for matrices A (the same as B).

        Returns
        -------
        sti
          Total order Sobol' indices, one for each dimensions.
        """
        nb_dim = self.masks_dim(masks)
        sampling_a, _, replication_c = self.split_abc(outputs, nb_design, nb_dim)

        mu_a = np.mean(sampling_a)
        var = np.sum([(v - mu_a)**2 for v in sampling_a]) / (len(sampling_a) - 1)

        stis = [
            (var - (1. / nb_design) * np.sum(sampling_a * replication_c[i]) + mu_a**2.0) / var
            for i in range(nb_dim)
        ]

        return self.post_process(stis, masks)


class JanonEstimator(SobolEstimator):
    """
    Janon estimator for total order Sobol' indices.

    Ref. Janon & al., Asymptotic normality and efficiency of two Sobol index estimators (2014)
    https://hal.inria.fr/hal-00665048v2/document
    """

    def __call__(self, masks, outputs, nb_design):
        """
        Compute the Sobol' total order indices according to the Janon algorithm.

        Parameters
        ----------
        masks
          Low resolution masks (before upsampling) used, one for each output.
        outputs
          Model outputs associated to each masks. One for each sample point of
          matrices A, B and C (in order).
        nb_design
          Number of points for matrices A (the same as B).

        Returns
        -------
        sti
          Total order Sobol' indices, one for each dimensions.
        """
        nb_dim = self.masks_dim(masks)
        sampling_a, _, replication_c = self.split_abc(outputs, nb_design, nb_dim)

        mu_ac = [(1. / nb_design) * np.sum(sampling_a + replication_c[i]) / 2.
                 for i in range(nb_dim)]
        var   = [(1. / (nb_design - 1.)) * np.sum(sampling_a**2. + replication_c[i]**2.) /
                  2. - mu_ac[i]**2. for i in range(nb_dim)]

        stis = [
            1. - ((1. / nb_design) * np.sum(sampling_a * replication_c[i]) - mu_ac[i]**2.) / var[i]
            for i in range(nb_dim)
        ]

        return self.post_process(stis, masks)


class GlenEstimator(SobolEstimator):
    """
    Glen-Isaacs estimator for total order Sobol' indices.

    Ref. Glen & al., Estimating Sobol sensitivity indices using correlations (2012)
    https://dl.acm.org/doi/abs/10.1016/j.envsoft.2012.03.014
    """

    def __call__(self, masks, outputs, nb_design):
        """
        Compute the Sobol' total order indices according to the Glen-Isaacs algorithm.

        Parameters
        ----------
        masks
          Low resolution masks (before upsampling) used, one for each output.
        outputs
          Model outputs associated to each masks. One for each sample point of
          matrices A, B and C (in order).
        nb_design
          Number of points for matrices A (the same as B).

        Returns
        -------
        sti
          Total order Sobol' indices, one for each dimensions.
        """
        nb_dim = self.masks_dim(masks)
        sampling_a, _, replication_c = self.split_abc(outputs, nb_design, nb_dim)

        mu_a = np.mean(sampling_a)
        mu_c = [np.mean(ci) for ci in replication_c]

        var_a = np.var(sampling_a)
        var_c = [np.var(ci) for ci in replication_c]

        stis = [
            1.0 - (1. / (nb_design - 1.) * np.sum((sampling_a - mu_a) *
                  (replication_c[i] - mu_c[i])) / (var_a * var_c[i])**0.5)
            for i in range(nb_dim)
        ]

        return self.post_process(stis, masks)


class SaltelliEstimator(SobolEstimator):
    """
    Saltelli estimator for total order Sobol' indices.

    Ref. Satelli & al., Global Sensitivity Analysis. The Primer.
    https://onlinelibrary.wiley.com/doi/book/10.1002/9780470725184
    """

    def __call__(self, masks, outputs, nb_design):
        """
        Compute the Sobol' total order indices according to the Saltelli algorithm.

        Parameters
        ----------
        masks
          Low resolution masks (before upsampling) used, one for each output.
        outputs
          Model outputs associated to each masks. One for each sample point of
          matrices A, B and C (in order).
        nb_design
          Number of points for matrices A (the same as B).

        Returns
        -------
        sti
          Total order Sobol' indices, one for each dimensions.
        """
        nb_dim = self.masks_dim(masks)
        sampling_a, _, replication_c = self.split_abc(outputs, nb_design, nb_dim)

        mu_a = np.mean(sampling_a)
        var = np.sum([(v - mu_a)**2 for v in sampling_a]) / (len(sampling_a) - 1)

        stis = [
            1. - ((1. / nb_design) * np.sum(sampling_a * replication_c[i]) - mu_a**2.) / var
            for i in range(nb_dim)
        ]

        return self.post_process(stis, masks)



import jax
import jaxopt
import jax.numpy as jnp
import numbers
import numpy as np
import scipy.sparse as sp
import time
import warnings
from math import sqrt
from typing import NamedTuple, Optional, Callable, Tuple

import tensorflow as tf
from jaxopt import base 
from typing import Optional
from dataclasses import dataclass
from sklearn.decomposition._nmf import _initialize_nmf as sklearn_initialize_nmf
from sklearn.decomposition import NMF as sk_NMF
from jaxopt._src import implicit_diff as idf

from ..utils import jax_l2


def nmf_value_and_gradcam(nmf, W, feature_model, x, concept_index, factoriseur, kkt_sol):
  x = tf.constant(x)
  x = x[None,:,:,:]
  with tf.GradientTape(persistent=True) as tape:  # persistent important !!

    tape.watch(x)
    gcA, A = feature_model(x)
    tape.watch(gcA)

    A = tf.transpose(A)

    # print(jnp.array(A).shape, jnp.array(W).shape)
    nnls_sol = nmf.init_params((jnp.array(A), jnp.array(W)))
    # print(nnls_sol.primal.shape)
    nnls_sol, backward_fun, state = jax.vjp(nmf.run,
                                            nnls_sol, (jnp.array(A), jnp.array(W)),
                                            has_aux=True)
    ua = tf.constant(nnls_sol.primal)
    tape.watch(ua)
    # print(ua.shape)
    #print(f"[{concept_index}] NNLS error={state.error} Iter Num={state.iter_num}")
    
    ui = ua[0,concept_index]

    with tape.stop_recording():  # avoid the tape to record itself
      [tangeant_vector_loss_fn] = tape.gradient([ui], [ua])

      # print(f"tangeant_vector_loss_fn.shape = {tangeant_vector_loss_fn.shape}")

      # Backward in Block n°2 [Jax]
      cotangeant_primal = jnp.array(tangeant_vector_loss_fn)
      cotangeant_dual   = jnp.zeros_like(nnls_sol.dual_ineq)
      cotangeant_vector_nmf = base.KKTSolution(
          primal = cotangeant_primal,
          dual_eq = None, dual_ineq = cotangeant_dual)  # gradient wrt dual variables is zero.
      # the cotangeant vector is now a tangeant vector
      tangeant_vector_kkt_sol, (tangeant_vector_a, tangeant_vector_dict) = backward_fun(cotangeant_vector_nmf)
      del tangeant_vector_kkt_sol  # unused
      del tangeant_vector_dict     # unused but we could compute it if we want

      # print(f"tangeant_vector_nmf.shape = {tangeant_vector_a.shape}")

    # Resume tape.
    # Backward in Block n°1 [Tf]
    cotangeant_vector_features = tf.constant(tangeant_vector_a)  # same shape as Y
    target = tf.reduce_sum(cotangeant_vector_features * A)  # scalar product
    # print(f"target value = {target}")

  tangeant_vector_x = tape.gradient(target, gcA)  # dui / dx
  gc_map = tf.reduce_sum(gcA * tf.reduce_mean(tangeant_vector_x, (1, 2))[:, None, None, :], -1)
  gc_map = tf.image.resize(tf.nn.relu(gc_map)[:, :, :, None], (224, 224), method='bicubic')
  return gc_map, ui


def nmf_gradcam(feature_model, x, factoriseur, kkt_sol):
  pairs = [nmf_value_and_gradcam(feature_model, x, i, factoriseur, kkt_sol) for i in range(factoriseur.rank)]
  grads = [grad for grad, _ in pairs]
  u = tf.stack([ui for _, ui in pairs])
  jacobian = tf.concat(grads, axis=0)
  return jacobian, u



import jax
import jaxopt
import jax.numpy as jnp
import numbers
import numpy as np
import scipy.sparse as sp
import time
import warnings
from math import sqrt
from typing import NamedTuple, Optional, Callable, Tuple

from jaxopt import base 
from typing import Optional
from dataclasses import dataclass
from sklearn.decomposition._nmf import _initialize_nmf as sklearn_initialize_nmf
from sklearn.decomposition import NMF as sk_NMF
from jaxopt._src import implicit_diff as idf

from .nnls import NNLS, NNLSState
from ..utils import jax_l2

def _make_nmf_kkt_optimality_fun():
  def obj_fun(primal_var, params_obj):
    H1, H2 = primal_var
    Y      = params_obj
    return 0.5 * jax_l2(Y - H1 @ H2.T)

  def ineq_fun(primal_var, params_ineq):
    H1, H2 = primal_var
    return -H1, -H2 # -H1 <= 0 and -H2 <= 0

  return idf.make_kkt_optimality_fun(obj_fun=obj_fun, eq_fun=None, ineq_fun=ineq_fun)


class NMFState(NamedTuple):
  """Named tuple containing state information.

  Attributes:
    iter_num: iteration number.
    error: error used as stop criterion, deduced from residuals. 
    nnls_state_1: NNLState associated to H1.
    nnls_state_2: NNLState associated to H2.
  """
  iter_num: int
  error: float
  nnls_state_1: NNLSState
  nnls_state_2: NNLSState


@dataclass(eq=False)
class NMF(base.IterativeSolver):
  rank: int
  maxiter: int = 100
  tol: float = 1e-3
  init: Optional[str] = None
  verbose: int = 0
  implicit_diff: bool = True
  implicit_diff_solve: Optional[Callable] = None
  jit: base.AutoOrBoolean = "auto"
  unroll: base.AutoOrBoolean = "auto"
  nnls: NNLS = NNLS()

  def _split_kkt_sol(self, params):
    H1, H2       = params.primal
    U1, U2       = params.dual_ineq
    kkt_sol_1    = base.KKTSolution(primal=H1, dual_eq=None, dual_ineq=U1)
    kkt_sol_2    = base.KKTSolution(primal=H2, dual_eq=None, dual_ineq=U2)
    return kkt_sol_1, kkt_sol_2

  def _merge_kkt_sol(self, kkt_sol_1, kkt_sol_2):
    primal = kkt_sol_1.primal, kkt_sol_2.primal
    dual_ineq = kkt_sol_1.dual_ineq, kkt_sol_2.dual_ineq
    return base.KKTSolution(primal=primal, dual_eq=None, dual_ineq=dual_ineq)

  def init_params(self,
                  params_obj: jnp.array,
                  params_eq: None = None,
                  params_ineq: None = None):
    Y = params_obj
    if self.init == 'bcd':
      sk_nmf = sk_NMF(n_components=self.rank)
      H1 = sk_nmf.fit_transform(Y)
      H2_T = sk_nmf.components_
    else:
      H1, H2_T = sklearn_initialize_nmf(Y, n_components=self.rank, init=self.init)
    H1, H2 = jnp.array(H1), jnp.array(H2_T.T)
    U1, U2 = jnp.zeros_like(H1), jnp.zeros_like(H2)
    params = base.KKTSolution(primal=(H1, H2),
                              dual_eq=None,
                              dual_ineq=(U1, U2))
    return params

  def init_state(self,
                 init_params: base.KKTSolution,
                 params_obj: jnp.array,
                 params_eq: None = None,
                 params_ineq: None = None):
    Y            = params_obj
    kkt_sol_1, kkt_sol_2 = self._split_kkt_sol(init_params)
    nnls_state_1 = self.nnls.init_state(kkt_sol_1, (Y.T, kkt_sol_2.primal))
    nnls_state_2 = self.nnls.init_state(kkt_sol_2, (Y,   kkt_sol_1.primal))
    return NMFState(
        iter_num=jnp.asarray(0, dtype=jnp.int32),
        error=jnp.asarray(jnp.inf),
        nnls_state_1=nnls_state_1,
        nnls_state_2=nnls_state_2,
    )

  def update(self, params, state, params_obj, params_eq, params_ineq):
    """Update state of NMF.
    
    n: number of rows
    m: number of columns
    k: rank of low rank factorization

    Args:
      params: KKTSolution tuple, with params.primal = (H1, H2),
              H1 of shape (n, k), and H2 of shape (m, k)
      state: NMFState object.
      params_obj: Y of shape (n, m).
      params_eq: None, present for signature purposes.
      params_ineq: None, present for signature purposes.

    Returns:
      pair params, 
    """
    H1, H2 = params.primal
    U1, U2 = params.dual_ineq
    Y = params_obj

    kkt_sol_1, kkt_sol_2 = self._split_kkt_sol(params)

    # Solve \|Y.T - H2 H1.T\| = \|Y - H1 H2.T\| for H1
    kkt_sol_1, nnls_state_1 = self.nnls.run(kkt_sol_1, (Y.T, kkt_sol_2.primal))

    # Solve \|Y - H1 H2.T\| for H2
    kkt_sol_2, nnls_state_2 = self.nnls.run(kkt_sol_2, (Y  , kkt_sol_1.primal))

    H1, H2 = kkt_sol_1.primal, kkt_sol_2.primal
    error = jax_l2(Y - H1 @ H2.T) / jax_l2(Y)

    next_params = self._merge_kkt_sol(kkt_sol_1, kkt_sol_2)

    next_state = NMFState(
        iter_num=state.iter_num+1,
        error=error,
        nnls_state_1=nnls_state_1,
        nnls_state_2=nnls_state_2,
    )
    
    return base.OptStep(next_params, next_state)

  def run(self,
          params: Optional[base.KKTSolution],
          params_obj: jnp.array,
          params_eq: None = None,
          params_ineq: None = None):
    if params is None:
      params = self.init_params(params_obj, params_eq, params_ineq)
    return super().run(params, params_obj, params_eq, params_ineq)

  def __post_init__(self):
    self.optimality_fun = _make_nmf_kkt_optimality_fun()


import jax
import jaxopt
import jax.numpy as jnp
from dataclasses import dataclass
from typing import NamedTuple, Optional, Callable, Tuple
from jaxopt._src import implicit_diff as idf
import jaxopt.base as base
from sklearn.decomposition import non_negative_factorization

from jaxopt.base import AutoOrBoolean, KKTSolution

from ..utils import jax_l2


class NNLSState(NamedTuple):
  """Named tuple containing state information.
  Attributes:
    iter_num: iteration number.
    error: error used as stop criterion, deduced from residuals.
    primal_residuals: relative residuals primal problem.
    dual_residuals: relative residuals dual problem.
    rho: step size in ADMM.
    H_bar: previous value of H_bar, useful for warm start.  
  """
  iter_num: int
  error: float
  primal_residuals: jnp.array
  dual_residuals: jnp.array
  rho: float
  H_bar: jnp.array


class NNLS(base.IterativeSolver):
  """
  Non-Negative Least Square solver for NMF with implicit differentiation
  """

  def __init__(self, max_iter: int = 1000, tolerance: float = 1e-4, 
               conjugate_gradient_tolerance : float = 1e-6, init_mode: str = "bcd", 
               implicit_diff : bool = True, implicit_diff_solver : Optional[Callable] = None, 
               jit : AutoOrBoolean = "auto", unroll : AutoOrBoolean = "auto"):
    self.max_iter = max_iter
    self.tolerance = tolerance
    self.conjugate_gradient_tolerance = conjugate_gradient_tolerance

    assert init_mode in ['zeros', 'bcd']
    self.init_mode = init_mode
    self.implicit_diff = implicit_diff
    self.implicit_diff_solver = implicit_diff_solver

    self.jit = jit
    self.unroll = unroll
    
    self.optimality_function = self.init_optimality_function()
  
  def init_optimality_function(self):
    """
    Instantiate the jaxopt optimality function for Karush–Kuhn–Tucker (KKT) conditions.
    """
    def objective_function(primal_var, params_obj):
      W = primal_var
      Y, U = params_obj
      
      residual = 0.5 * jax_l2(Y - U @ W.T)
      return residual 
    
    def inequality_constraint(primal_var, params_ineq):
      U = primal_var
      # as the inequality are defined such that X <= 0 is imposed
      # and we want positivity, then -X <= 0 mean that X is positive
      return -U

    return idf.make_kkt_optimality_fun(obj_fun=objective_function, 
                                       eq_fun=None, 
                                       ineq_fun=inequality_constraint)
  
  def init_params(self,
                  params_obj: Tuple[jnp.array, jnp.array]):
    """
    Initialize the U,W and prepare the Karush–Kuhn–Tucker (KKT) solution state.
    """
    Y, W = params_obj
    # as mentioned in the paper, `r` is the lowering dimension (number of concept)
    # and `n` the number of points.
    n, r = Y.shape[1], W.shape[1]

    if self.init == 'zeros':
      U = jnp.zeros((n, r))
    
    elif self.init == 'bcd':
      # use sklearn to find a good initialization point
      start_U, start_W = non_negative_factorization(Y.T, H=W.T, update_H=False)
      # sklearn's W and H are inverted: it is not a bug !
      U = jnp.array(start_U.T)
      print("U shape", U.shape)
      assert U.shape == (n, r)
    
    kkt_solution = KKTSolution(primal=U,
                               dual_eq=None,
                               dual_ineq=U)
    return kkt_solution

  def init_state(self, 
                 init_params,
                 params_obj: Tuple[jnp.array, jnp.array]):
    """
    Initialize the Non-negative Least Square state
    """
    Y, W = params_obj
    r = W.shape[1]
    
    rho = jnp.asarray(jnp.sum(W**2) / r)
    
    state = NNLSState(
        iter_num = jnp.asarray(0, dtype=jnp.int32),
        error = jnp.asarray(jnp.inf),
        primal_residuals = jnp.asarray(jnp.inf),
        dual_residuals = jnp.asarray(jnp.inf),
        rho = rho,
        H_bar = init_params.primal.T,
    )

    return state

  def _compute_H_bar(self, H, G, F, U, rho, H_bar):
    # solution to argmin_H | H - H_bar + U|_2 + delta(H)
    # G is PSD so G + rho * I is PSD => Conjugate Gradient can be used
    def matmul(vec):
      return jnp.dot(G, vec) + rho * vec
    right_member = F + rho * (H + U).T
    H_bar, _ = jax.scipy.sparse.linalg.cg(matmul, right_member,
                                          x0=H_bar, tol=self.conjugate_gradient_tolerance)
    return H_bar
  
  def update(self, params, state, params_obj, params_eq, params_ineq):
    """
    Execute one step and update the state of the Non-negative Least Square.

    with n: number of rows (points),
         p: number of columns (features)
         r: dimension of low rank factorization (number of concepts)

    params
      KKTSolution tuple, with params.primal = H and H of shape (m, k)
    state
      NNLSState object.
    params_obj
      Pair (Y, W), Y of shape (n, m) and W of shape (n, k)
    params_eq
      None, present for signature purposes.
    params_ineq
      None, present for signature purposes.

    Returns:
      pair params.
    """

    Y, W = params_obj
    F = W.T @ Y
    G = W.T @ W  # PSD matrix.
    H, U = params.primal, params.dual_ineq

    # ADMM first inner problem.
    H_bar = self._compute_H_bar(H, G, F, U, state.rho, state.H_bar)

    # ADMM second inner problem.
    H = jax.nn.relu(H_bar.T - U)

    # Gradient ascent step on dual variables.
    U = U + H - H_bar.T

    primal_residuals = jax_l2(H - H_bar.T) / jax_l2(H)
    dual_residuals = jax_l2(H - params.primal) / jax_l2(U)
    error = jnp.maximum(primal_residuals, dual_residuals)

    next_params = base.KKTSolution(primal=H, dual_eq=None, dual_ineq=U)

    next_state = NNLSState(
        iter_num=state.iter_num+1,
        error=error,
        primal_residuals=primal_residuals,
        dual_residuals=dual_residuals,
        rho=state.rho,
        H_bar=H_bar,
    )

    return base.OptStep(next_params, next_state)

  def run(self,
          params: Optional[base.KKTSolution],
          params_obj: Tuple[jnp.array, jnp.array],
          params_eq: None = None,
          params_ineq: None = None):
    if params is None:
      params = self.init_params(params_obj, params_eq, params_ineq)
    return super().run(params, params_obj, params_eq, params_ineq)


"""
Sampling methods for replicated designs
"""

from abc import ABC, abstractmethod

import numpy as np
import tensorflow as tf
import scipy


class Sampler(ABC):
    """
    Base class for replicated design sampling.
    """

    @staticmethod
    def build_replicated_design(sampling_a, sampling_b):
        """
        Build the replicated design matrix C using A & B

        Parameters
        ----------
        sampling_a
          The masks values for the sampling matrix A.
        sampling_b
          The masks values for the sampling matrix B.

        Returns
        -------
        replication_c
          The new replicated design matrix C generated from A & B.
        """
        replication_c = np.array([sampling_a.copy() for _ in range(sampling_a.shape[-1])])
        for i in range(len(replication_c)):
            replication_c[i, :, i] = sampling_b[:, i]

        replication_c = replication_c.reshape((-1, sampling_a.shape[-1]))

        return replication_c

    @abstractmethod
    def __call__(self, dimension, nb_design):
        raise NotImplementedError()


class ScipySampler(Sampler):
    """
    Base class based on Scipy qmc module for replicated design sampling.
    """

    def __init__(self):
        try:
            self.qmc = scipy.stats.qmc # pylint: disable=E1101
        except AttributeError as err:
            raise ModuleNotFoundError("Xplique need scipy>=1.7 to use this sampling.") from err


class TFSobolSequence(Sampler):
    """
    Tensorflow Sobol LP tau sequence sampler.
    Ref. I. M. Sobol., The distribution of points in a cube and the accurate evaluation of
    integrals (1967).
    https://www.sciencedirect.com/science/article/abs/pii/0041555367901449
    """

    def __call__(self, dimension, nb_design):
        sampling_ab = tf.math.sobol_sample(dimension*2, nb_design, dtype=tf.float32).numpy()
        sampling_a, sampling_b = sampling_ab[:, :dimension], sampling_ab[:, dimension:]
        replicated_c = self.build_replicated_design(sampling_a, sampling_b)

        return np.concatenate([sampling_a, sampling_b, replicated_c], 0)


class ScipySobolSequence(ScipySampler):
    """
    Scipy Sobol LP tau sequence sampler.

    Ref. I. M. Sobol., The distribution of points in a cube and the accurate evaluation of
    integrals (1967).
    https://www.sciencedirect.com/science/article/abs/pii/0041555367901449
    """

    def __call__(self, dimension, nb_design):
        sampler = self.qmc.Sobol(dimension*2, scramble=False)
        sampling_ab = sampler.random(nb_design).astype(np.float32)
        sampling_a, sampling_b = sampling_ab[:, :dimension], sampling_ab[:, dimension:]
        replicated_c = self.build_replicated_design(sampling_a, sampling_b)

        return np.concatenate([sampling_a, sampling_b, replicated_c], 0)


class HaltonSequence(ScipySampler):
    """
    Halton sequence sampler.

    Ref. J.H. Halton., On the efficiency of certain quasi-random sequences of points in evaluating
    multi-dimensional integrals (1960).
    https://link.springer.com/article/10.1007/BF01386213
    """

    def __call__(self, dimension, nb_design):
        sampler = self.qmc.Halton(dimension*2, scramble=False)
        sampling_ab = sampler.random(nb_design).astype(np.float32)
        sampling_a, sampling_b = sampling_ab[:, :dimension], sampling_ab[:, dimension:]
        replicated_c = self.build_replicated_design(sampling_a, sampling_b)

        return np.concatenate([sampling_a, sampling_b, replicated_c], 0)


class LHSampler(ScipySampler):
    """
    Latin hypercube sampler.

    Ref. Mckay & al., A Comparison of Three Methods for Selecting Values of Input Variables in the
    Analysis of Output from a Computer Code (1979).
    https://www.jstor.org/stable/1268522
    """

    def __call__(self, dimension, nb_design):
        sampler = self.qmc.LatinHypercube(dimension*2)
        sampling_ab = sampler.random(nb_design).astype(np.float32)
        sampling_a, sampling_b = sampling_ab[:, :dimension], sampling_ab[:, dimension:]
        replicated_c = self.build_replicated_design(sampling_a, sampling_b)

        return np.concatenate([sampling_a, sampling_b, replicated_c], 0)


from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as fh:
    README = fh.read()

setup(
    name="Craft-xai",
    version="0.0.1",
    description="Automatic Concept Extraction with CRAFT",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Thomas FEL, Agustin PICARD, Louis Béthune, Thibaut BOISSIN",
    author_email="thomas_fel@brown.edu",
    license="MIT",
    install_requires=['numpy', 'scikit-learn', 'scikit-image',
                      'matplotlib', 'scipy', 'opencv-python'],
    extras_require={
        "tests": ["pytest", "pylint"],
        "docs": ["mkdocs", "mkdocs-material", "numkdoc"],
    },
    packages=find_packages(),
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)


"""
CRAFT Module for Tensorflow
"""

from abc import ABC, abstractmethod
from typing import Union, Tuple, List, Callable, Dict, Optional, Any

import tensorflow as tf
import numpy as np
from sklearn.decomposition import NMF
from sklearn.exceptions import NotFittedError

from .sobol.sampler import HaltonSequence
from .sobol.estimators import JansenEstimator


class BaseConceptExtractor(ABC):
    """
    Base class for concept extraction models.

    Parameters
    ----------
    input_to_latent : Callable
        The first part of the model taking an input and returning
        positive activations, g(.) in the original paper.
    latent_to_logit : Callable
        The second part of the model taking activation and returning
        logits, h(.) in the original paper.
    number_of_concepts : int
        The number of concepts to extract.
    batch_size : int, optional
        The batch size to use during training and prediction. Default is 64.

    """

    def __init__(self, input_to_latent : Callable,
                       latent_to_logit : Optional[Callable] = None,
                       number_of_concepts: int = 20,
                       batch_size: int = 64):

        # sanity checks
        assert(number_of_concepts > 0), "number_of_concepts must be greater than 0"
        assert(batch_size > 0), "batch_size must be greater than 0"
        assert(callable(input_to_latent)), "input_to_latent must be a callable function"

        self.input_to_latent = input_to_latent
        self.latent_to_logit = latent_to_logit
        self.number_of_concepts = number_of_concepts
        self.batch_size = batch_size

    @abstractmethod
    def fit(self, inputs):
        """
        Fit the CAVs to the input data.

        Parameters
        ----------
        inputs : array-like
            The input data to fit the model on.

        Returns
        -------
        tuple
            A tuple containing the input data and the matrices (U, W) that factorize the data.

        """
        raise NotImplementedError

    @abstractmethod
    def transform(self, inputs):
        """
        Transform the input data into a concepts embedding.

        Parameters
        ----------
        inputs : array-like
            The input data to transform.

        Returns
        -------
        array-like
            The transformed embedding of the input data.

        """
        raise NotImplementedError


class Craft(BaseConceptExtractor):
    """
    Class Implementing the CRAFT Concept Extraction Mechanism.

    Parameters
    ----------
    input_to_latent : Callable
        The first part of the model taking an input and returning
        positive activations, g(.) in the original paper.
    latent_to_logit : Callable
        The second part of the model taking activation and returning
        logits, h(.) in the original paper.
    number_of_concepts : int
        The number of concepts to extract.
    batch_size : int, optional
        The batch size to use during training and prediction. Default is 64.
    patch_size : int, optional
        The size of the patches to extract from the input data. Default is 64.
    """

    def __init__(self, input_to_latent : Callable,
                       latent_to_logit : Optional[Callable] = None,
                       number_of_concepts: int = 20,
                       batch_size: int = 64,
                       patch_size: int = 64):
        super().__init__(input_to_latent, latent_to_logit, number_of_concepts, batch_size)

        self.patch_size = patch_size
        self.activation_shape = None

    def fit(self, inputs : np.ndarray):
        """
        Fit the Craft model to the input data.

        Parameters
        ----------
        inputs : np.ndarray
            Input data of shape (n_samples, height, width, channels).
            (x1, x2, ..., xn) in the paper.

        Returns
        -------
        (X, U, W)
            A tuple containing the crops (X in the paper),
            the concepts values (U) and the concepts basis (W).
        """
        # extract patches from the input data
        strides = int(self.patch_size * 0.80)
        patches = tf.image.extract_patches(images=inputs,
                                           sizes=[1, self.patch_size, self.patch_size, 1],
                                           strides=[1, strides, strides, 1],
                                           rates=[1, 1, 1, 1],
                                           padding='VALID')
        patches = tf.reshape(patches, (-1, self.patch_size, self.patch_size, 3))

        # encode the patches and obtain the activations
        input_width, input_height = inputs.shape[1], inputs.shape[2]
        activations = self.input_to_latent.predict(tf.image.resize(patches, (input_width, input_height), method="bicubic"),
                                                  batch_size=self.batch_size,
                                                  verbose=False)

        assert np.min(activations) >= 0.0, "Activations must be positive."

        # if the activations have shape (n_samples, height, width, n_channels),
        # apply average pooling
        if len(activations.shape) == 4:
            activations = tf.reduce_mean(activations, axis=(1, 2))

        # apply NMF to the activations to obtain matrices U and W
        reducer = NMF(n_components=self.number_of_concepts, alpha_W=1e-2)
        U = tf.cast(reducer.fit_transform(tf.nn.relu(activations)), tf.float32)
        W = tf.cast(reducer.components_, tf.float32)

        # store the factorizer and W as attributes of the Craft instance
        self.reducer = reducer
        self.W = tf.cast(W, tf.float32)

        return patches, U, W


    def check_if_fitted(self):
        """Checks if the factorization model has been fitted to input data.

        Raises
        ------
        NotFittedError
            If the factorization model has not been fitted to input data.
        """

        if not hasattr(self, 'reducer'):
            raise NotFittedError("The factorization model has not been fitted to input data yet.")

    def transform(self, inputs : np.ndarray, activations : np.ndarray = None):
        """Transforms the inputs data into its concept representation.

        Parameters
        ----------
        inputs : numpy array or Tensor
            The input data to be transformed.
        activations: numpy array or Tensor, optional
            Pre-computed activations of the input data. If not provided, the activations
            will be computed using the input_to_latent model.

        Returns
        -------
        U : Tensor
            The concept value (U) of the inputs.
        """
        self.check_if_fitted()

        activations = self.input_to_latent.predict(inputs, batch_size=self.batch_size,
                                                   verbose=False)

        if len(activations.shape) == 4:
            original_shape = activations.shape[:-1]
            activations = np.reshape(activations, (-1, activations.shape[-1]))
        else:
            original_shape = (len(activations),)

        W_dtype = self.reducer.components_.dtype
        U = self.reducer.transform(np.array(activations, dtype=W_dtype))
        U = np.reshape(U, (*original_shape, U.shape[-1]))

        return tf.cast(U, tf.float32)

    def estimate_importance(self, inputs, class_id, nb_design = 32):
        """
        Estimates the importance of each concept for a given class.

        Parameters
        ----------
        inputs : numpy array or Tensor
            The input data to be transformed.
        class_id : int
            The class id to estimate the importance for.
        nb_design : int, optional
            The number of design to use for the importance estimation. Default is 32.

        Returns
        -------
        importances : list
            The Sobol total index (importance score) for each concept.

        """
        self.check_if_fitted()

        U = self.transform(inputs)

        masks = HaltonSequence()(self.number_of_concepts, nb_design = nb_design)
        estimator = JansenEstimator()

        importances = []

        if len(U.shape) == 2:
            # apply the original method of the paper

            for u in U:
                u_perturbated = u[None, :] * masks
                a_perturbated = u_perturbated @ self.W

                y_pred = self.latent_to_logit.predict(a_perturbated, batch_size=self.batch_size,
                                                      verbose=False)
                y_pred = y_pred[:, class_id]

                stis = estimator(masks, y_pred, nb_design)

                importances.append(stis)

        elif len(U.shape) == 4:
            # apply a re-parameterization trick and use mask on all localization for a given
            # concept id to estimate sobol indices
            for u in U:
                u_perturbated = u[None, :] * masks[:, None, None, :]
                a_perturbated = np.reshape(u_perturbated, (-1, u.shape[-1])) @ self.W
                a_perturbated = np.reshape(a_perturbated, (len(masks), U.shape[1], U.shape[2], -1))

                y_pred = self.latent_to_logit.predict(a_perturbated, batch_size=self.batch_size,
                                                      verbose=False)
                y_pred = y_pred[:, class_id]

                stis = estimator(masks, y_pred, nb_design)

                importances.append(stis)

        return np.mean(importances, 0)
